/*
 * SPDX-License-Identifier: ISC
 *
 * Copyright (c) 2018 Todd C. Miller <Todd.Miller@sudo.ws>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef SUDOERS_STRLIST_H
#define SUDOERS_STRLIST_H

/*
 * Simple string list with optional reference count.
 */
struct sudoers_string {
    STAILQ_ENTRY(sudoers_string) entries;
    char *str;
};
struct sudoers_str_list {
    struct sudoers_string *stqh_first;
    struct sudoers_string **stqh_last;
    unsigned int refcnt;
};

struct sudoers_str_list *str_list_alloc(void);
void str_list_free(void *v);
struct sudoers_string *sudoers_string_alloc(const char *s);
void sudoers_string_free(struct sudoers_string *ls);

#endif /* SUDOERS_STRLIST_H */
